/*
 * Use of this source code is governed by the below BSD-2-Clause license
 *
 * Copyright (c) 2020, Ayappan Perumal. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	@(#)getifaddrs_aix.c	1.0
 */

#include <string.h>
#include <sys/ioctl.h>
#include <stdlib.h>
#include <net/if.h>
#include <LightGBM/ifaddrs.h>
#include <unistd.h>

int getifaddrs(struct ifaddrs **ifap)
{
    char *buf;
    int fd4, fd6, fd;
    unsigned i;
    int ifconf_size;
    int ifsize;
    sa_family_t ifr_af;
    struct ifconf ifc;
    struct ifreq *ifrp=NULL;
    struct ifreq ifr4;
    struct in6_ifreq ifr6;
    struct ifaddrs *curif, *lastif;


    if ((fd4 = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
        return -1;
    }
    if ((fd6 = socket(AF_INET6, SOCK_DGRAM, 0)) < 0) {
        close(fd4);
        return -1;
    }

    if (ioctl(fd4, SIOCGSIZIFCONF, &ifconf_size) < 0)
        goto fail;

    ifc.ifc_len = ifconf_size;
    buf = (char*)calloc(1, ifconf_size);
    ifc.ifc_buf = buf;

    if (ioctl(fd4, SIOCGIFCONF, &ifc) < 0) {
        free(buf);
        close(fd4);
        close(fd6);
        return -1;
    }

    ifrp = ifc.ifc_req;

    i = ifc.ifc_len;

    *ifap = NULL;
    lastif = NULL;

    /* Loop through all the available interfaces */
    while (i > 0)
    {
		ifsize = sizeof(struct ifreq);

		if (ifrp->ifr_addr.sa_len > sizeof(ifrp->ifr_ifru)) {
			ifsize += ifrp->ifr_addr.sa_len - sizeof(ifrp->ifr_ifru);
		}

		strncpy(ifr4.ifr_name, ifrp->ifr_name, sizeof(ifr4.ifr_name));
		strncpy(ifr6.ifr_name, ifrp->ifr_name, sizeof(ifr6.ifr_name));
		ifr_af = ifrp->ifr_addr.sa_family;

		if (ifr_af != AF_INET && ifr_af != AF_INET6)
			goto next;

		fd = (ifr_af == AF_INET ? fd4 : fd6);

        /* Get interface flags */
		if (ioctl(fd, SIOCGIFFLAGS, &ifr4) < 0)
			goto fail;

		if (!(ifr4.ifr_flags & IFF_UP)) {
			goto next;
		}

		curif = (struct ifaddrs*)calloc(1, sizeof (struct ifaddrs));
		if (curif == NULL)
			goto fail;

		if (lastif != NULL) {
			lastif->ifa_next = curif;
		} else {
			*ifap = curif;
		}
		lastif = curif;

		curif->ifa_flags = ifr4.ifr_flags;
		if ((curif->ifa_name = strdup(ifrp->ifr_name)) == NULL)
			goto fail;

        /* Store interface address */
		curif->ifa_addr = (struct sockaddr*)calloc(1, sizeof(struct sockaddr));
        if (curif->ifa_addr == NULL)
            goto fail;
        memcpy(curif->ifa_addr, &ifrp->ifr_addr, sizeof(struct sockaddr));

		/* Get netmask */
		if (ifr_af == AF_INET) {
			if (ioctl(fd, SIOCGIFNETMASK, &ifr4) < 0) {
				goto fail;
			}
            curif->ifa_netmask = (struct sockaddr*)calloc(1, sizeof(struct sockaddr));
            if (curif->ifa_netmask == NULL)
                goto fail;
            memcpy(curif->ifa_netmask, &ifr4.ifr_addr, sizeof(struct sockaddr));
		} else {
			if (ioctl(fd, SIOCGIFNETMASK6, &ifr6) < 0) {
				goto fail;
			}
            curif->ifa_netmask = (struct sockaddr*)calloc(1, sizeof(struct sockaddr));
            if (curif->ifa_netmask == NULL)
                goto fail;
            memcpy(curif->ifa_netmask, &ifr6.ifr_Addr, sizeof(struct sockaddr));
		}

		/* Get destination address for P2P interface */
		if (curif->ifa_flags & IFF_POINTOPOINT) {
			if (ifr_af == AF_INET) {
				if (ioctl(fd, SIOCGIFDSTADDR, &ifr4) < 0) {
					goto fail;
                }
                curif->ifa_dstaddr = (struct sockaddr*)calloc(1, sizeof(struct sockaddr));
                if (curif->ifa_dstaddr == NULL)
                    goto fail;
                memcpy(curif->ifa_dstaddr, &ifr4.ifr_addr, sizeof(struct sockaddr));
			} else {
				if (ioctl(fd, SIOCGIFDSTADDR6, &ifr6) < 0) {
					goto fail;
                }
                curif->ifa_dstaddr = (struct sockaddr*)calloc(1, sizeof(struct sockaddr));
                if (curif->ifa_dstaddr == NULL)
                    goto fail;
                memcpy(curif->ifa_dstaddr, &ifr6.ifr_Addr, sizeof(struct sockaddr));
			}

        /* Get broadcast address (only for IPv4) */
		} else if ((curif->ifa_flags & IFF_BROADCAST) && (ifr_af == AF_INET)) {
			if (ioctl(fd, SIOCGIFBRDADDR, &ifr4) < 0) {
				goto fail;
            }
            curif->ifa_broadaddr = (struct sockaddr*)calloc(1, sizeof(struct sockaddr));
            if (curif->ifa_broadaddr == NULL)
                goto fail;
            memcpy(curif->ifa_broadaddr, &ifr4.ifr_addr, sizeof(struct sockaddr));
		}
next:
        ifrp = (struct ifreq*)((char*)ifrp + ifsize);
		i -= ifsize;
	}

	free(buf);
	close(fd4);
	close(fd6);
	return 0;
fail:
	free(buf);
	freeifaddrs(*ifap);
	*ifap = NULL;
	close(fd4);
	close(fd6);
	return -1;
}

void freeifaddrs(struct ifaddrs *ifa)
{
    struct ifaddrs *curif;

    while (ifa != NULL) {
        curif = ifa; 
        ifa = ifa->ifa_next;
        free(curif->ifa_name);
        free(curif->ifa_addr);
        free(curif->ifa_netmask);
        free(curif->ifa_dstaddr);
        free(curif);
    }
}
